@@ -1,43 +1,60 @@
-/* $Xorg: imDefIm.c,v 1.5 2000/08/17 19:45:12 cpqbld Exp $ */
+/*
+ * Copyright (c) 1990, 1991, 1992, Oracle and/or its affiliates.
+ *
+ * Permission is hereby granted, free of charge, to any person obtaining a
+ * copy of this software and associated documentation files (the "Software"),
+ * to deal in the Software without restriction, including without limitation
+ * the rights to use, copy, modify, merge, publish, distribute, sublicense,
+ * and/or sell copies of the Software, and to permit persons to whom the
+ * Software is furnished to do so, subject to the following conditions:
+ *
+ * The above copyright notice and this permission notice (including the next
+ * paragraph) shall be included in all copies or substantial portions of the
+ * Software.
+ *
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+ * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+ * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
+ * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+ * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
+ * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
+ * DEALINGS IN THE SOFTWARE.
+ */
+
 /******************************************************************
-         Copyright 1990, 1991, 1992 by Sun Microsystems, Inc.
          Copyright 1992, 1993, 1994 by FUJITSU LIMITED
          Copyright 1993, 1994 by Sony Corporation
 
-Permission to use, copy, modify, distribute, and sell this software and its
-documentation for any purpose is hereby granted without fee, provided that
-the above copyright notice appear in all copies and that both that copyright
-notice and this permission notice appear in supporting documentation, and
-that the name of Sun Microsystems, Inc., FUJITSU LIMITED and Sony
-Corporation not be used in advertising or publicity pertaining to
-distribution of the software without specific, written prior permission.
-Sun Microsystems, Inc., FUJITSU LIMITED and Sony Corporation makes no
-representations about the suitability of this software for any purpose.  It
-is provided "as is" without express or implied warranty. 
+Permission to use, copy, modify, and distribute this software and its
+documentation for any purpose and without fee is hereby granted,
+provided that the above copyright notice appear in all copies and that
+both that copyright notice and this permission notice appear in
+supporting documentation, and that the names of Digital, FUJITSU
+LIMITED and Sony Corporation not be used in advertising or publicity
+pertaining to distribution of the software without specific, written
+prior permission.
 
-Sun Microsystems Inc., FUJITSU LIMITED AND SONY CORPORATION DISCLAIMS ALL
-WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
-MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL Sun Microsystems, Inc.,
-FUJITSU LIMITED AND SONY CORPORATION BE LIABLE FOR ANY SPECIAL, INDIRECT OR
-CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
-DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
-TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
-OF THIS SOFTWARE. 
+DIGITAL, FUJITSU LIMITED AND SONY CORPORATION DISCLAIMS ALL WARRANTIES
+WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
+MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL DIGITAL, FUJITSU LIMITED
+AND SONY CORPORATION BE LIABLE FOR ANY SPECIAL, INDIRECT OR
+CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
+USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
+OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
+PERFORMANCE OF THIS SOFTWARE.
 
   Author: Hideki Hiura (hhiura@Sun.COM) Sun Microsystems, Inc.
-          Takashi Fujiwara     FUJITSU LIMITED 
+          Takashi Fujiwara     FUJITSU LIMITED
                                fujiwara@a80.tech.yk.fujitsu.co.jp
           Makoto Wakamatsu     Sony Corporation
                                makoto@sm.sony.co.jp
 
 ******************************************************************/
-/* $XFree86: xc/lib/X11/imDefIm.c,v 1.15 2003/08/22 19:27:24 eich Exp $ */
 
 #ifdef HAVE_CONFIG_H
 #include <config.h>
 #endif
 #include <X11/Xatom.h>
-#define NEED_EVENTS
 #include "Xlibint.h"
 #include "Xlcint.h"
 #include "XlcPublic.h"
@@ -45,8 +62,11 @@ OF THIS SOFTWARE.
 #include "XimTrInt.h"
 #include "Ximint.h"
 
+#include <limits.h>
+#include <stdlib.h>
+#include <strings.h>
 
-Public int
+int
 _XimCheckDataSize(
     XPointer	 buf,
     int		 len)
@@ -58,7 +78,7 @@ _XimCheckDataSize(
     return  buf_s[1];
 }
 
-Public void
+void
 _XimSetHeader(
     XPointer	 buf,
     CARD8	 major_opcode,
@@ -76,7 +96,7 @@ _XimSetHeader(
     return;
 }
 
-Public char
+char
 _XimGetMyEndian(void)
 {
     CARD16	 test_card = 1;
@@ -87,7 +107,7 @@ _XimGetMyEndian(void)
 	return BIGENDIAN;
 }
 
-Private Bool
+static Bool
 _XimCheckServerName(
     Xim		   im,
     char	  *str)
@@ -109,7 +129,7 @@ _XimCheckServerName(
 
     if(strncmp(str, XIM_SERVER_CATEGORY, category_len))
 	return False;
- 
+
     pp = &str[category_len];
 
     for(;;) {
@@ -125,7 +145,7 @@ _XimCheckServerName(
     return True;
 }
 
-Private char *
+static char *
 _XimCheckLocaleName(
     Xim		   im,
     char	  *address,
@@ -145,7 +165,7 @@ _XimCheckLocaleName(
 
     if(strncmp(address, XIM_LOCAL_CATEGORY, category_len))
 	return (char*)NULL;
- 
+
     pp = &address[category_len];
 
     for(;;) {
@@ -165,11 +185,11 @@ _XimCheckLocaleName(
     return (char *)NULL;
 }
 
-Private Bool
+static Bool
 _XimCheckTransport(
     char	  *address,
     int		   address_len,
-    char	  *transport,
+    const char	  *transport,
     int		   len,
     char	 **trans_addr)
 {
@@ -182,7 +202,7 @@ _XimCheckTransport(
 
     if(strncmp(address, XIM_TRANSPORT_CATEGORY, category_len))
 	return False;
- 
+
     pp = &address[category_len];
 
     for(;;) {
@@ -209,7 +229,7 @@ _XimCheckTransport(
     return True;
 }
 
-Private Bool
+static Bool
 _CheckSNEvent(
     Display		*display,
     XEvent		*xevent,
@@ -223,7 +243,7 @@ _CheckSNEvent(
     return False;
 }
 
-Private Bool
+static Bool
 _XimGetSelectionNotify(
     Display		 *display,
     Window		  window,
@@ -252,12 +272,12 @@ _XimGetSelectionNotify(
     return True;
 }
 
-Private Bool
+static Bool
 _XimPreConnectionIM(
     Xim			 im,
     Atom		 selection)
 {
-    Display		*display = im->core.display; 
+    Display		*display = im->core.display;
     Atom		 locales, transport;
     char		*address;
     XLCd		 lcd;
@@ -313,9 +333,7 @@ _XimPreConnectionIM(
 	else
 	    goto Error;
     }
-    if( (locale_name[3] = Xmalloc(llen+1)) != NULL )
-	strcpy( locale_name[3], language );
-    else
+    if( (locale_name[3] = strdup(language)) == NULL )
 	goto Error;
     if((locales = XInternAtom(display, XIM_LOCALES, True)) == (Atom)None)
 	goto Error;
@@ -367,14 +385,14 @@ Error:
 	if( locale_name[i] != NULL )
 	    XFree( locale_name[i] );
     XDestroyWindow(display, window);
-    return False; 
+    return False;
 }
 
-Private Bool
+static Bool
 _XimPreConnect(
     Xim		    im)
 {
-    Display	   *display = im->core.display; 
+    Display	   *display = im->core.display;
     Atom	    imserver;
     Atom	    actual_type;
     int		    actual_format;
@@ -384,12 +402,13 @@ _XimPreConnect(
     Atom	   *atoms;
     Window	    im_window = 0;
     register int    i;
+    const char	   *env_enable_fabricated_order;
 
     if((imserver = XInternAtom(display, XIM_SERVERS, True)) == (Atom)None)
 	return False;
 
     if(XGetWindowProperty(display, RootWindow(display, 0),
-			imserver, 0L, 1000000L, False, XA_ATOM, &actual_type, 
+			imserver, 0L, 1000000L, False, XA_ATOM, &actual_type,
 			&actual_format, &nitems, &bytes_after,
 			&prop_return) != Success)
 	return False;
@@ -414,10 +433,20 @@ _XimPreConnect(
 	return False;
 
     im->private.proto.im_window = im_window;
+    im->private.proto.fabricated_serial = 0;
+    im->private.proto.fabricated_time = 0;
+    im->private.proto.enable_fabricated_order = True;
+    env_enable_fabricated_order = getenv("LIBX11_ENABLE_FABRICATED_ORDER");
+    if (env_enable_fabricated_order && *env_enable_fabricated_order) {
+	if (!strncasecmp(env_enable_fabricated_order, "0", 2) ||
+	    !strncasecmp(env_enable_fabricated_order, "false", 6)) {
+	    im->private.proto.enable_fabricated_order = False;
+	}
+    }
     return True;
 }
 
-Private Bool
+static Bool
 _XimGetAuthProtocolNames(
     Xim		 im,
     CARD16	*buf,
@@ -435,7 +464,7 @@ _XimGetAuthProtocolNames(
     return True;
 }
 
-Private Bool
+static Bool
 _XimSetAuthReplyData(
     Xim		 im,
     XPointer	 buf,
@@ -448,7 +477,7 @@ _XimSetAuthReplyData(
     return True;
 }
 
-Private Bool
+static Bool
 _XimSetAuthNextData(
     Xim		 im,
     XPointer	 buf,
@@ -461,7 +490,7 @@ _XimSetAuthNextData(
     return True;
 }
 
-Private Bool
+static Bool
 _XimSetAuthRequiredData(
     Xim		 im,
     XPointer	 buf,
@@ -474,7 +503,7 @@ _XimSetAuthRequiredData(
     return True;
 }
 
-Private Bool
+static Bool
 _XimCheckAuthSetupData(
     Xim		 im,
     XPointer	 buf)
@@ -485,7 +514,7 @@ _XimCheckAuthSetupData(
     return True;
 }
 
-Private Bool
+static Bool
 _XimCheckAuthNextData(
     Xim		 im,
     XPointer	 buf)
@@ -500,7 +529,7 @@ _XimCheckAuthNextData(
 #define	GOOD_AUTH	1
 #define	BAD_AUTH	0
 
-Private int
+static int
 _XimClientAuthCheck(
     Xim		 im,
     XPointer	 buf)
@@ -511,7 +540,7 @@ _XimClientAuthCheck(
     return NO_MORE_AUTH;
 }
 
-Private void
+static void
 _XimAuthNG(
     Xim		 im)
 {
@@ -525,7 +554,7 @@ _XimAuthNG(
     return;
 }
 
-Private	Bool
+static Bool
 _XimAllRecv(
     Xim		 im,
     INT16	 len,
@@ -538,7 +567,7 @@ _XimAllRecv(
 #define	CLIENT_WAIT1		1
 #define	CLIENT_WAIT2		2
 
-Private Bool
+static Bool
 _XimConnection(
     Xim		 im)
 {
@@ -599,7 +628,7 @@ _XimConnection(
 		preply = reply;
 	    } else {
 		buf_size = len;
-		preply = (XPointer)Xmalloc(buf_size);
+		preply = Xmalloc(buf_size);
 		ret_code = _XimRead(im, &len, preply, buf_size, _XimAllRecv, 0);
 		if(ret_code != XIM_TRUE) {
 		    Xfree(preply);
@@ -684,7 +713,7 @@ _XimConnection(
 	}
     }
 
-    if (!( buf_s[0] == im->private.proto.protocol_major_version 
+    if (!( buf_s[0] == im->private.proto.protocol_major_version
         && buf_s[1] == im->private.proto.protocol_minor_version)) {
 	if(reply != preply)
 	    Xfree(preply);
@@ -699,7 +728,7 @@ _XimConnection(
     return True;
 }
 
-Private	Bool
+static Bool
 _XimDisconnectCheck(
     Xim		 im,
     INT16	 len,
@@ -718,7 +747,7 @@ _XimDisconnectCheck(
     return False;
 }
 
-Private Bool
+static Bool
 _XimDisconnect(
     Xim		 im)
 {
@@ -742,7 +771,7 @@ _XimDisconnect(
 	if(ret_code == XIM_OVERFLOW) {
 	    if(len > 0) {
 		buf_size = len;
-		preply = (XPointer)Xmalloc(buf_size);
+		preply = Xmalloc(buf_size);
 		ret_code = _XimRead(im, &len, preply, buf_size,
 						 _XimDisconnectCheck, 0);
 		Xfree(preply);
@@ -758,7 +787,7 @@ _XimDisconnect(
     return True;
 }
 
-Private	Bool
+static Bool
 _XimOpenCheck(
     Xim		 im,
     INT16	 len,
@@ -777,7 +806,7 @@ _XimOpenCheck(
     return False;
 }
 
-Private Bool
+static Bool
 _XimOpen(
     Xim			 im)
 {
@@ -792,12 +821,16 @@ _XimOpen(
     int			 buf_size;
     int			 ret_code;
     char		*locale_name;
+    size_t		 locale_len;
 
     locale_name = im->private.proto.locale_name;
-    len = strlen(locale_name);
-    buf_b[0] = (BYTE)len;			   /* length of locale name */
-    (void)strcpy((char *)&buf_b[1], locale_name);  /* locale name */
-    len += sizeof(BYTE);			   /* sizeof length */
+    locale_len = strlen(locale_name);
+    if (locale_len > UCHAR_MAX)
+      return False;
+    memset(buf32, 0, sizeof(buf32));
+    buf_b[0] = (BYTE)locale_len;		/* length of locale name */
+    memcpy(&buf_b[1], locale_name, locale_len);	   /* locale name */
+    len = (INT16)(locale_len + sizeof(BYTE));	   /* sizeof length */
     XIM_SET_PAD(buf_b, len);			   /* pad */
 
     _XimSetHeader((XPointer)buf, XIM_OPEN, 0, &len);
@@ -814,7 +847,7 @@ _XimOpen(
 	    preply = reply;
 	} else {
 	    buf_size = len;
-	    preply = (XPointer)Xmalloc(buf_size);
+	    preply = Xmalloc(buf_size);
 	    ret_code = _XimRead(im, &len, preply, buf_size,
 					_XimOpenCheck, 0);
 	    if(ret_code != XIM_TRUE) {
@@ -854,7 +887,7 @@ _XimOpen(
     _XimSetIMMode(im->private.proto.im_inner_resources,
 				im->private.proto.im_num_inner_resources);
 
-    /* Transport Callbak */
+    /* Transport Callback */
     _XimRegProtoIntrCallback(im, XIM_SET_EVENT_MASK, 0,
 				 _XimSetEventMaskCallback, (XPointer)im);
     _XimRegProtoIntrCallback(im, XIM_FORWARD_EVENT, 0,
@@ -873,7 +906,7 @@ _XimOpen(
     return True;
 }
 
-Private	Bool
+static Bool
 _XimCloseCheck(
     Xim		 im,
     INT16	 len,
@@ -897,7 +930,7 @@ _XimCloseCheck(
     return False;
 }
 
-Private Bool
+static Bool
 _XimClose(
     Xim		 im)
 {
@@ -918,7 +951,7 @@ _XimClose(
     buf_s[1] = 0;				/* unused */
     len = sizeof(CARD16)			/* sizeof imid */
         + sizeof(CARD16);			/* sizeof unused */
-  
+
     _XimSetHeader((XPointer)buf, XIM_CLOSE, 0, &len);
     if (!(_XimWrite(im, len, (XPointer)buf)))
 	return False;
@@ -933,7 +966,7 @@ _XimClose(
 	    preply = reply;
 	} else {
 	    buf_size = len;
-	    preply = (XPointer)Xmalloc(buf_size);
+	    preply = Xmalloc(buf_size);
 	    ret_code = _XimRead(im, &len, preply, buf_size, _XimCloseCheck, 0);
 	    if(ret_code != XIM_TRUE) {
 		Xfree(preply);
@@ -955,39 +988,35 @@ _XimClose(
     return True;
 }
 
-Public void
+void
 _XimProtoIMFree(
     Xim		  im)
 {
     /* XIMPrivateRec */
-    if (im->private.proto.im_onkeylist) {
-	Xfree(im->private.proto.im_onkeylist);
-	im->private.proto.im_onkeylist = NULL;
-    }
-    if (im->private.proto.im_offkeylist) {
-	Xfree(im->private.proto.im_offkeylist);
-	im->private.proto.im_offkeylist = NULL;
-    }
+
+    Xfree(im->private.proto.im_onkeylist);
+    im->private.proto.im_onkeylist = NULL;
+
+    Xfree(im->private.proto.im_offkeylist);
+    im->private.proto.im_offkeylist = NULL;
+
     if (im->private.proto.intrproto) {
-	_XimFreeProtoIntrCallback(im);
-	im->private.proto.intrproto = NULL;
-    }
-    if (im->private.proto.im_inner_resources) {
-	Xfree(im->private.proto.im_inner_resources);
-	im->private.proto.im_inner_resources = NULL;
-    }
-    if (im->private.proto.ic_inner_resources) {
-	Xfree(im->private.proto.ic_inner_resources);
-	im->private.proto.ic_inner_resources = NULL;
-    }
-    if (im->private.proto.hold_data) {
-	Xfree(im->private.proto.hold_data);
-	im->private.proto.hold_data = NULL;
-    }
-    if (im->private.proto.locale_name) {
-	Xfree(im->private.proto.locale_name);
-	im->private.proto.locale_name = NULL;
+      _XimFreeProtoIntrCallback(im);
+      im->private.proto.intrproto = NULL;
     }
+
+    Xfree(im->private.proto.im_inner_resources);
+    im->private.proto.im_inner_resources = NULL;
+
+    Xfree(im->private.proto.ic_inner_resources);
+    im->private.proto.ic_inner_resources = NULL;
+
+    Xfree(im->private.proto.hold_data);
+    im->private.proto.hold_data = NULL;
+
+    Xfree(im->private.proto.locale_name);
+    im->private.proto.locale_name = NULL;
+
     if (im->private.proto.ctom_conv) {
 	_XlcCloseConverter(im->private.proto.ctom_conv);
 	im->private.proto.ctom_conv = NULL;
@@ -1027,53 +1056,42 @@ _XimProtoIMFree(
     }
 #endif /* XIM_CONNECTABLE */
 
-    if (im->private.proto.saved_imvalues) {
-        Xfree(im->private.proto.saved_imvalues);
-        im->private.proto.saved_imvalues = NULL;
-    }
-    if (im->private.proto.default_styles) {
-	Xfree(im->private.proto.default_styles);
-	im->private.proto.default_styles = NULL;
-    }
+    Xfree(im->private.proto.saved_imvalues);
+    im->private.proto.saved_imvalues = NULL;
+
+    Xfree(im->private.proto.default_styles);
+    im->private.proto.default_styles = NULL;
 
     /* core */
-    if (im->core.res_name) {
-        Xfree(im->core.res_name);
-	im->core.res_name = NULL;
-    }
-    if (im->core.res_class) {
-        Xfree(im->core.res_class);
-	im->core.res_class = NULL;
-    }
-    if (im->core.im_values_list) {
-	Xfree(im->core.im_values_list);
-	im->core.im_values_list = NULL;
-    }
-    if (im->core.ic_values_list) {
-	Xfree(im->core.ic_values_list);
-	im->core.ic_values_list = NULL;
-    }
-    if (im->core.im_name) {
-	Xfree(im->core.im_name);
-	im->core.im_name = NULL;
-    }
-    if (im->core.styles) {
-	Xfree(im->core.styles);
-	im->core.styles = NULL;
-    }
-    if (im->core.im_resources) {
-        Xfree(im->core.im_resources);
-	im->core.im_resources = NULL;
-    }
-    if (im->core.ic_resources) {
-        Xfree(im->core.ic_resources);
-	im->core.ic_resources = NULL;
-    }
+
+    Xfree(im->core.res_name);
+    im->core.res_name = NULL;
+
+    Xfree(im->core.res_class);
+    im->core.res_class = NULL;
+
+    Xfree(im->core.im_values_list);
+    im->core.im_values_list = NULL;
+
+    Xfree(im->core.ic_values_list);
+    im->core.ic_values_list = NULL;
+
+    Xfree(im->core.im_name);
+    im->core.im_name = NULL;
+
+    Xfree(im->core.styles);
+    im->core.styles = NULL;
+
+    Xfree(im->core.im_resources);
+    im->core.im_resources = NULL;
+
+    Xfree(im->core.ic_resources);
+    im->core.ic_resources = NULL;
 
     return;
 }
 
-Private Status
+static Status
 _XimProtoCloseIM(
     XIM		 xim)
 {
@@ -1088,10 +1106,10 @@ _XimProtoCloseIM(
 	next = ic->core.next;
 #ifdef XIM_CONNECTABLE
 	if (!(!IS_SERVER_CONNECTED(im) && IS_RECONNECTABLE(im))) {
-	    Xfree ((char *) ic);
+	    Xfree (ic);
 	}
 #else
-	Xfree ((char *) ic);
+	Xfree (ic);
 #endif /* XIM_CONNECTABLE */
 	ic = next;
     }
@@ -1121,7 +1139,7 @@ _XimProtoCloseIM(
 }
 
 #ifdef XIM_CONNECTABLE
-Private Bool
+static Bool
 _XimCheckIMQuarkList(
     XrmQuark		*quark_list,
     int			 num_quark,
@@ -1137,7 +1155,7 @@ _XimCheckIMQuarkList(
     return False;
 }
 
-Private Bool
+static Bool
 _XimSaveIMValues(
     Xim			 im,
     XIMArg		*arg)
@@ -1156,7 +1174,7 @@ _XimSaveIMValues(
 	    if (_XimCheckIMQuarkList(quark_list, num_quark, quark)) {
 		continue;
 	    }
-	    if (!(tmp = (XrmQuark *)Xrealloc(quark_list,
+	    if (!(tmp = Xrealloc(quark_list,
 				(sizeof(XrmQuark) * (num_quark + 1))))) {
 		im->private.proto.saved_imvalues = quark_list;
 		im->private.proto.num_saved_imvalues = num_quark;
@@ -1173,7 +1191,7 @@ _XimSaveIMValues(
 
     for (p = arg, n = 0; p && p->name; p++, n++);
 
-    if (!(quark_list = (XrmQuark *)Xmalloc(sizeof(XrmQuark) * n))) {
+    if (!(quark_list = Xmalloc(sizeof(XrmQuark) * n))) {
 	return False;
     }
 
@@ -1186,7 +1204,7 @@ _XimSaveIMValues(
     return True;
 }
 
-Private char *
+static char *
 _XimDelayModeSetIMValues(
     Xim			 im,
     XIMArg		*arg)
@@ -1204,7 +1222,7 @@ _XimDelayModeSetIMValues(
 }
 #endif /* XIM_CONNECTABLE */
 
-Private Bool
+static Bool
 _XimSetIMValuesCheck(
     Xim          im,
     INT16        len,
@@ -1228,7 +1246,7 @@ _XimSetIMValuesCheck(
     return False;
 }
 
-Private char *
+static char *
 _XimProtoSetIMValues(
     XIM			 xim,
     XIMArg		*arg)
@@ -1272,6 +1290,7 @@ _XimProtoSetIMValues(
 #endif /* XIM_CONNECTABLE */
 
     _XimGetCurrentIMValues(im, &im_values);
+    memset(tmp_buf, 0, sizeof(tmp_buf32));
     buf = tmp_buf;
     buf_size = XIM_HEADER_SIZE + sizeof(CARD16) + sizeof(INT16);
     data_len = BUFSIZE - buf_size;
@@ -1282,8 +1301,6 @@ _XimProtoSetIMValues(
 	if ((name = _XimEncodeIMATTRIBUTE(im, im->core.im_resources,
 		im->core.im_num_resources, arg, &arg_ret, data, data_len,
 		&ret_len, (XPointer)&im_values, XIM_SETIMVALUES))) {
-	    if (buf != tmp_buf)
-		Xfree(buf);
 	    break;
 	}
 
@@ -1294,23 +1311,27 @@ _XimProtoSetIMValues(
 
 	buf_size += ret_len;
 	if (buf == tmp_buf) {
-	    if (!(tmp = (char *)Xmalloc(buf_size + data_len))) {
+	    if (!(tmp = Xcalloc(buf_size + data_len, 1))) {
 		return arg->name;
 	    }
 	    memcpy(tmp, buf, buf_size);
 	    buf = tmp;
 	} else {
-	    if (!(tmp = (char *)Xrealloc(buf, (buf_size + data_len)))) {
+	    if (!(tmp = Xrealloc(buf, (buf_size + data_len)))) {
 		Xfree(buf);
 		return arg->name;
 	    }
+            memset(&tmp[buf_size], 0, data_len);
 	    buf = tmp;
 	}
     }
     _XimSetCurrentIMValues(im, &im_values);
 
-    if (!total)
-	return (char *)NULL;
+    if (!total) {
+        if (buf != tmp_buf)
+	    Xfree(buf);
+        return (char *)NULL;
+    }
 
     buf_s = (CARD16 *)&buf[XIM_HEADER_SIZE];
     buf_s[0] = im->private.proto.imid;
@@ -1336,7 +1357,7 @@ _XimProtoSetIMValues(
 	    preply = reply;
 	} else {
 	    buf_size = (int)len;
-	    preply = (XPointer)Xmalloc(buf_size);
+	    preply = Xmalloc(buf_size);
 	    ret_code = _XimRead(im, &len, reply, buf_size,
 					_XimSetIMValuesCheck, 0);
 	    if(ret_code != XIM_TRUE) {
@@ -1360,7 +1381,7 @@ _XimProtoSetIMValues(
 }
 
 #ifdef XIM_CONNECTABLE
-Private char *
+static char *
 _XimDelayModeGetIMValues(
     Xim			 im,
     XIMArg		*arg)
@@ -1373,7 +1394,7 @@ _XimDelayModeGetIMValues(
 }
 #endif /* XIM_CONNECTABLE */
 
-Private Bool
+static Bool
 _XimGetIMValuesCheck(
     Xim          im,
     INT16        len,
@@ -1397,7 +1418,7 @@ _XimGetIMValuesCheck(
     return False;
 }
 
-Private char *
+static char *
 _XimProtoGetIMValues(
     XIM			 xim,
     XIMArg		*arg)
@@ -1445,7 +1466,7 @@ _XimProtoGetIMValues(
 	     + sizeof(INT16)
 	     + XIM_PAD(buf_size);
 
-    if (!(buf = (CARD8 *)Xmalloc(buf_size)))
+    if (!(buf = Xcalloc(buf_size, 1)))
 	return arg->name;
     buf_s = (CARD16 *)&buf[XIM_HEADER_SIZE];
 
@@ -1477,7 +1498,7 @@ _XimProtoGetIMValues(
 		preply = reply;
 	    } else {
 		buf_size = len;
-		preply = (XPointer)Xmalloc(buf_size);
+		preply = Xmalloc(buf_size);
 		ret_code = _XimRead(im, &len, preply, buf_size,
 						_XimGetIMValuesCheck, 0);
 		if(ret_code != XIM_TRUE) {
@@ -1509,7 +1530,7 @@ _XimProtoGetIMValues(
 	return makeid_name;
 }
 
-Private XIMMethodsRec     im_methods = {
+static XIMMethodsRec     im_methods = {
     _XimProtoCloseIM,           /* close */
     _XimProtoSetIMValues,       /* set_values */
     _XimProtoGetIMValues,       /* get_values */
@@ -1519,7 +1540,7 @@ Private XIMMethodsRec     im_methods = {
     _Ximctstoutf8		/* ctstoutf8 */
 };
 
-Private Bool
+static Bool
 _XimSetEncodingByName(
     Xim		  im,
     char	**buf,
@@ -1539,20 +1560,20 @@ _XimSetEncodingByName(
     encoding_len = strlen(encoding);
     compound_len = strlen("COMPOUND_TEXT");
     *len = encoding_len + sizeof(BYTE) + compound_len + sizeof(BYTE);
-    if (!(ret = (BYTE *)Xmalloc(*len))) {
+    if (!(ret = Xmalloc(*len))) {
 	return False;
     }
     *buf = (char *)ret;
 
     ret[0] = (BYTE)encoding_len;
-    (void)strncpy((char *)&ret[1], encoding, encoding_len);
+    memcpy(&ret[1], encoding, encoding_len);
     ret += (encoding_len + sizeof(BYTE));
     ret[0] = (BYTE)compound_len;
-    (void)strncpy((char *)&ret[1], "COMPOUND_TEXT", compound_len);
+    memcpy(&ret[1], "COMPOUND_TEXT", compound_len);
     return True;
 }
 
-Private Bool
+static Bool
 _XimSetEncodingByDetail(
     Xim		 im,
     char	**buf,
@@ -1563,7 +1584,7 @@ _XimSetEncodingByDetail(
     return True;
 }
 
-Private Bool
+static Bool
 _XimGetEncoding(
     Xim		 im,
     CARD16	*buf,
@@ -1652,7 +1673,7 @@ _XimGetEncoding(
     return True;
 }
 
-Private	Bool
+static Bool
 _XimEncodingNegoCheck(
     Xim		 im,
     INT16	 len,
@@ -1676,7 +1697,7 @@ _XimEncodingNegoCheck(
     return False;
 }
 
-Private Bool
+static Bool
 _XimEncodingNegotiation(
     Xim		 im)
 {
@@ -1696,11 +1717,8 @@ _XimEncodingNegotiation(
     if (!(_XimSetEncodingByName(im, &name_ptr, &name_len)))
 	return False;
 
-    if (!(_XimSetEncodingByDetail(im, &detail_ptr, &detail_len))) {
-	if (name_ptr)
-	    Xfree(name_ptr);
-	return False;
-    }
+    if (!(_XimSetEncodingByDetail(im, &detail_ptr, &detail_len)))
+	goto free_name_ptr;
 
     len = sizeof(CARD16)
 	+ sizeof(INT16)
@@ -1710,13 +1728,9 @@ _XimEncodingNegotiation(
 	+ sizeof(CARD16)
 	+ detail_len;
 
-    if (!(buf = (CARD8 *)Xmalloc(XIM_HEADER_SIZE + len))) {
-	if (name_ptr)
-	    Xfree(name_ptr);
-	if (detail_ptr)
-	    Xfree(detail_ptr);
-	return False;
-    }
+    if (!(buf = Xcalloc(XIM_HEADER_SIZE + len, 1)))
+	goto free_detail_ptr;
+
     buf_s = (CARD16 *)&buf[XIM_HEADER_SIZE];
 
     buf_s[0] = im->private.proto.imid;
@@ -1733,7 +1747,7 @@ _XimEncodingNegotiation(
     _XimSetHeader((XPointer)buf, XIM_ENCODING_NEGOTIATION, 0, &len);
     if (!(_XimWrite(im, len, (XPointer)buf))) {
 	Xfree(buf);
-	return False;
+	goto free_detail_ptr;
     }
     _XimFlush(im);
     Xfree(buf);
@@ -1747,43 +1761,47 @@ _XimEncodingNegotiation(
 	    preply = reply;
 	} else {
 	    buf_size = len;
-	    preply = (XPointer)Xmalloc(buf_size);
+	    preply = Xmalloc(buf_size);
 	    ret_code = _XimRead(im, &len, preply, buf_size,
 					_XimEncodingNegoCheck, 0);
-	    if(ret_code != XIM_TRUE) {
-		Xfree(preply);
-		return False;
-	    }
+	    if(ret_code != XIM_TRUE)
+		goto free_preply;
 	}
     } else
-	return False;
+	goto free_detail_ptr;
     buf_s = (CARD16 *)((char *)preply + XIM_HEADER_SIZE);
     if (*((CARD8 *)preply) == XIM_ERROR) {
 	_XimProcError(im, 0, (XPointer)&buf_s[3]);
-	if(reply != preply)
-	    Xfree(preply);
-	return False;
+	goto free_preply;
     }
 
     if (!(_XimGetEncoding(im, &buf_s[1], name_ptr, name_len,
-						detail_ptr, detail_len))) {
-	if(reply != preply)
-	    Xfree(preply);
-	return False;
-    }
-    if (name_ptr)
-	Xfree(name_ptr);
-    if (detail_ptr)
-	Xfree(detail_ptr);
+						detail_ptr, detail_len)))
+	goto free_preply;
+
+    Xfree(name_ptr);
+    Xfree(detail_ptr);
 
     if(reply != preply)
 	Xfree(preply);
 
     return True;
+
+free_preply:
+    if (reply != preply)
+	Xfree(preply);
+
+free_detail_ptr:
+    Xfree(detail_ptr);
+
+free_name_ptr:
+    Xfree(name_ptr);
+
+    return False;
 }
 
 #ifdef XIM_CONNECTABLE
-Private Bool
+static Bool
 _XimSendSavedIMValues(
     Xim			 im)
 {
@@ -1806,6 +1824,7 @@ _XimSendSavedIMValues(
     int			 ret_code;
 
     _XimGetCurrentIMValues(im, &im_values);
+    memset(tmp_buf, 0, sizeof(tmp_buf32));
     buf = tmp_buf;
     buf_size = XIM_HEADER_SIZE + sizeof(CARD16) + sizeof(INT16);
     data_len = BUFSIZE - buf_size;
@@ -1828,16 +1847,17 @@ _XimSendSavedIMValues(
 
 	buf_size += ret_len;
 	if (buf == tmp_buf) {
-	    if (!(tmp = (char *)Xmalloc(buf_size + data_len))) {
+	    if (!(tmp = Xcalloc(buf_size + data_len, 1))) {
 		return False;
 	    }
 	    memcpy(tmp, buf, buf_size);
 	    buf = tmp;
 	} else {
-	    if (!(tmp = (char *)Xrealloc(buf, (buf_size + data_len)))) {
+	    if (!(tmp = Xrealloc(buf, (buf_size + data_len)))) {
 		Xfree(buf);
 		return False;
 	    }
+            memset(&tmp[buf_size], 0, data_len);
 	    buf = tmp;
 	}
     }
@@ -1869,7 +1889,7 @@ _XimSendSavedIMValues(
 	    preply = reply;
 	} else {
 	    buf_size = (int)len;
-	    preply = (XPointer)Xmalloc(buf_size);
+	    preply = Xmalloc(buf_size);
 	    ret_code = _XimRead(im, &len, reply, buf_size,
 					_XimSetIMValuesCheck, 0);
     	    if(ret_code != XIM_TRUE) {
@@ -1893,7 +1913,7 @@ _XimSendSavedIMValues(
     return True;
 }
 
-Private void
+static void
 _XimDelayModeIMFree(
     Xim		 im)
 {
@@ -1916,13 +1936,13 @@ _XimDelayModeIMFree(
     return;
 }
 
-Public Bool
+Bool
 _XimConnectServer(
     Xim		 im)
 {
     Xim		 save_im;
 
-    if (!(save_im = (Xim)Xmalloc(sizeof(XimRec))))
+    if (!(save_im = Xmalloc(sizeof(XimRec))))
 	return False;
     memcpy((char *)save_im, (char *)im, sizeof(XimRec));
 
@@ -1940,7 +1960,7 @@ _XimConnectServer(
     return False;
 }
 
-Public Bool
+Bool
 _XimDelayModeSetAttr(
     Xim			 im)
 {
@@ -1972,7 +1992,7 @@ _XimDelayModeSetAttr(
     return True;
 }
 
-Private Bool
+static Bool
 _XimReconnectModeSetAttr(
     Xim			 im)
 {
@@ -1999,7 +2019,7 @@ _XimReconnectModeSetAttr(
 }
 #endif /* XIM_CONNECTABLE */
 
-Public Bool
+Bool
 _XimProtoOpenIM(
     Xim		 im)
 {
